@@ -13,7 +13,9 @@ __author__ = "John Westbrook"
 __email__ = "jwest@rcsb.rutgers.edu"
 __license__ = "Apache 2.0"
 
+import hashlib
 import logging
+import os
 
 from rcsb.exdb.utils.ObjectExtractor import ObjectExtractor
 from rcsb.exdb.utils.ObjectUpdater import ObjectUpdater
@@ -21,6 +23,7 @@ from rcsb.utils.chem.ChemCompIndexProvider import ChemCompIndexProvider
 from rcsb.utils.chem.ChemCompSearchIndexProvider import ChemCompSearchIndexProvider
 from rcsb.utils.chemref.PubChemUtils import PubChemUtils, ChemicalIdentifier
 from rcsb.utils.io.IoUtil import getObjSize
+from rcsb.utils.io.MarshalUtil import MarshalUtil
 from rcsb.utils.io.TimeUtil import TimeUtil
 from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil
 
@@ -44,9 +47,9 @@ class ReferenceUpdateWorker(object):
             self.__lookupD.setdefault(ccId, []).append(sD)
         self.__databaseName = "pubchem_exdb"
         self.__refDataCollectionName = "reference_entry"
-        self.__matchDataCollectionName = "reference_match"
+        self.__matchIndexCollectionName = "reference_match_index"
         self.__createCollections(self.__databaseName, self.__refDataCollectionName, indexAttributeNames=["rcsb_id", "rcsb_last_update"])
-        self.__createCollections(self.__databaseName, self.__matchDataCollectionName, indexAttributeNames=["rcsb_id", "rcsb_last_update"])
+        self.__createCollections(self.__databaseName, self.__matchIndexCollectionName, indexAttributeNames=["rcsb_id", "rcsb_last_update"])
 
     def __genChemIdList(self, ccId):
         chemIdList = []
@@ -58,7 +61,7 @@ class ReferenceUpdateWorker(object):
                 elif "smiles" in sD:
                     idType = "smiles"
                     descr = sD["smiles"]
-                chemIdList.append(ChemicalIdentifier(idCode=ccId, identifierSource=sD["build-type"], identifierType=idType, identifier=descr))
+                chemIdList.append(ChemicalIdentifier(idCode=ccId, identifierSource=sD["build-type"], identifierType=idType, identifier=descr, indexName=sD["name"]))
         return chemIdList
 
     def updateList(self, dataList, procName, optionsD, workingDir):
@@ -69,8 +72,9 @@ class ReferenceUpdateWorker(object):
                     "rcsb_id" : "local reference ID (ccid|bird)", << LOCAL CANNONICAL ID (e.g. ATP, PRD_000100)
                     "rcsb_last_update" : ISODate("2020-04-08T16:26:47.993+0000"),
                     "matched_ids" : [
-                        {"matched_id":  "<external reference ID code>", "search_id_type" : "oe-smiles", "search_id_source": "model-xyz"},
-                        {"matched_id":  "<external reference ID code>", "search_id_type": ... , "search_id_source": ... }
+                        {"matched_id":  "<external reference ID code>", "search_id_type" : "oe-smiles", "search_id_source": "model-xyz",
+                                        'source_index_name': <>, 'source_inchikey': <>, 'source_smiles': <>},
+                        {"matched_id":  "<external reference ID code>", "search_id_type": ... , "search_id_source": ... , ...}
                         ]                          ]
                     },
                 }
@@ -81,8 +85,6 @@ class ReferenceUpdateWorker(object):
                     "rcsb_last_update" : ISODate("2020-04-08T16:26:48.025+0000"),
                 }
                 #
-                ChemicalIdentifier(idCode=uId, identifierSource=sD["build-type"], identifierType=idType, identifier=descr))
-                #
         """
         _ = workingDir
         chunkSize = optionsD.get("chunkSize", 50)
@@ -115,7 +117,26 @@ class ReferenceUpdateWorker(object):
                         if ok and refDL:
                             for tD in refDL:
                                 pcId = tD["cid"]
-                                mL.append({"matched_id": pcId, "search_id_type": chemId.identifierType, "search_id_source": chemId.identifierSource})
+                                inchiKey = (
+                                    self.__searchIdxD[chemId.indexName]["inchi-key"]
+                                    if chemId.indexName in self.__searchIdxD and "inchi-key" in self.__searchIdxD[chemId.indexName]
+                                    else None
+                                )
+                                smiles = (
+                                    self.__searchIdxD[chemId.indexName]["smiles"]
+                                    if chemId.indexName in self.__searchIdxD and "smiles" in self.__searchIdxD[chemId.indexName]
+                                    else None
+                                )
+                                mL.append(
+                                    {
+                                        "matched_id": pcId,
+                                        "search_id_type": chemId.identifierType,
+                                        "search_id_source": chemId.identifierSource,
+                                        "source_index_name": chemId.indexName,
+                                        "source_smiles": smiles,
+                                        "source_inchikey": inchiKey,
+                                    }
+                                )
                                 tD.update({"rcsb_id": pcId, "rcsb_last_update": timeS})
                                 tDL.append(tD)
                     #
@@ -127,15 +148,15 @@ class ReferenceUpdateWorker(object):
                     #
                     tIdxDL.append(tIdxD)
                 # --
-                self.__updateReferenceData(self.__databaseName, self.__matchDataCollectionName, tIdxDL)
-                self.__updateReferenceData(self.__databaseName, self.__refDataCollectionName, tDL)
+                self.__updateObjectStore(self.__databaseName, self.__matchIndexCollectionName, tIdxDL)
+                self.__updateObjectStore(self.__databaseName, self.__refDataCollectionName, tDL)
         except Exception as e:
             logger.exception("Failing %s for %d data items %s", procName, len(dataList), str(e))
         logger.info("%s dataList length %d success length %d rst1 %d rst2 %d", procName, len(dataList), len(successList), len(retList1), len(retList2))
         #
         return successList, emptyList, emptyList, diagList
 
-    def __updateReferenceData(self, databaseName, collectionName, objDL):
+    def __updateObjectStore(self, databaseName, collectionName, objDL):
         updateDL = []
         for objD in objDL:
             try:
@@ -162,7 +183,7 @@ class PubChemCacheProvider(object):
 
     """
 
-    def __init__(self, cfgOb, chunkSize=100, expireDays=0, numProc=1, fetchLimit=None, **kwargs):
+    def __init__(self, cfgOb, chunkSize=100, expireDays=0, numProc=1, fetchLimit=None, updateOnReload=True, retryUnmatched=False, **kwargs):
         self.__cfgOb = cfgOb
         #
         self.__chunkSize = chunkSize
@@ -171,9 +192,11 @@ class PubChemCacheProvider(object):
         #
         self.__databaseName = "pubchem_exdb"
         self.__refDataCollectionName = "reference_entry"
-        self.__matchDataCollectionName = "reference_match"
+        self.__matchIndexCollectionName = "reference_match_index"
         #
-        self.__matchD, self.__refD = self.__reload(expireDays, fetchLimit, **kwargs)
+        self.__matchD, self.__refD, self.__ccIdxP, self.__ccsidxP = self.__reload(
+            expireDays=expireDays, fetchLimit=fetchLimit, updateOnReload=updateOnReload, retryUnmatched=retryUnmatched, **kwargs
+        )
 
     def getMatchInfo(self):
         return self.__matchD
@@ -217,7 +240,7 @@ class PubChemCacheProvider(object):
             tU = TimeUtil()
             tS = tU.getTimestamp(useUtc=True, before={"days": expireDays})
             selectD.update({"matched_ids": {"$exists": True}})
-        matchD = self.__getReferenceData(self.__databaseName, self.__matchDataCollectionName, selectD=selectD if selectD else None)
+        matchD = self.__getReferenceData(self.__databaseName, self.__matchIndexCollectionName, selectD=selectD if selectD else None)
         return sorted(matchD.keys())
 
     # --
@@ -227,19 +250,113 @@ class PubChemCacheProvider(object):
             self.__cfgOb, databaseName=databaseName, collectionName=collectionName, keyAttribute="rcsb_id", uniqueAttributes=["rcsb_id"], selectionQuery=selectD,
         )
         docCount = obEx.getCount()
-        logger.debug("Reference data match count %d", docCount)
+        logger.info("Reference data match count %d", docCount)
         objD = obEx.getObjects()
         return objD
 
-    def __reload(self, expireDays, fetchLimit, **kwargs):
+    def transformIndex(self, exportPath=None):
         """[summary]
 
-        Args:
-            expireDays ([type]): [description]
-
         Returns:
             [type]: [description]
-            Match Object -
+
+
+                Example match index entry:
+                {
+                    "_id" : ObjectId("5e8dfb49eab967a0483a0472"),
+                    "rcsb_id" : "local reference ID (ccid|bird)", << LOCAL CANNONICAL ID (e.g. ATP, PRD_000100)
+                    "rcsb_last_update" : ISODate("2020-04-08T16:26:47.993+0000"),
+                    "matched_ids" : [
+                        {"matched_id":  "<external reference ID code>", "search_id_type" : "oe-smiles", "search_id_source": "model-xyz",
+                                        'source_index_name': <>, 'source_inchikey': <>, 'source_smiles': <>},
+                        {"matched_id":  "<external reference ID code>", "search_id_type": ... , "search_id_source": ... , ...}
+                        ]                          ]
+                    },
+                }
+        """
+        #
+        oL = []
+
+        if exportPath:
+            fp = os.path.join(exportPath, "match_data_orig.json")
+            mU = MarshalUtil(workPath=exportPath)
+            mU.doExport(fp, self.__matchD, fmt="json", indent=3)
+        #
+        mapD = {}
+        sIdxD = self.__ccsidxP.getIndex()
+        for sId, idxD in sIdxD.items():
+            ccId = sId.split("|")[0]
+            ss = idxD["build-type"].split("|")[0]
+            mId = ccId + "|" + ss if ss != "model-xyz" else ccId
+            mapD[mId] = sId
+
+        for ccId, mD in self.__matchD.items():
+            del mD["_id"]
+            #
+            if "matched_ids" in mD:
+                qL = []
+                for sD in mD["matched_ids"]:
+                    # lookup the information in the prior index
+                    sSource = sD["search_id_source"].split("|")[0]
+                    mId = ccId + "|" + sSource if sSource != "model-xyz" else ccId
+                    if mId in mapD:
+                        siD = sIdxD[mapD[mId]]
+                        smiles = siD["smiles"]
+                        qualifier = hashlib.sha256(smiles.encode("utf-8")).hexdigest()
+                        inchiKey = siD["inchi-key"]
+                        sourceIndexName = ccId + "|" + qualifier if sSource != "model-xyz" else ccId
+                        sD["source_smiles"] = smiles
+                        sD["source_index_name"] = sourceIndexName
+                        sD["source_inchikey"] = inchiKey
+                        sD["search_id_source"] = "tautomer|model-xyz" if sSource.startswith("tautomer") else sSource
+                        sD["search_id_source"] = "unique-protomer|model-xyz" if sD["search_id_source"].startswith("unique-") else sD["search_id_source"]
+                        qL.append(sD)
+                mD["matched_ids"] = qL
+            #
+            oL.append(mD)
+        logger.info("Index object length (%d)", len(oL))
+        if exportPath:
+            fp = os.path.join(exportPath, "match_index_object_list.json")
+            mU = MarshalUtil(workPath=exportPath)
+            mU.doExport(fp, oL, fmt="json", indent=3)
+        #
+        self.__createCollections(self.__databaseName, self.__matchIndexCollectionName, indexAttributeNames=["rcsb_id", "rcsb_last_update"])
+        self.__updateObjectStore(self.__databaseName, self.__matchIndexCollectionName, oL)
+        #
+
+        #
+        return True
+
+    def __createCollections(self, databaseName, collectionName, indexAttributeNames=None):
+        obUpd = ObjectUpdater(self.__cfgOb)
+        ok = obUpd.createCollection(databaseName, collectionName, indexAttributeNames=indexAttributeNames, checkExists=True, bsonSchema=None)
+        return ok
+
+    def __updateObjectStore(self, databaseName, collectionName, objDL):
+        updateDL = []
+        for objD in objDL:
+            try:
+                selectD = {"rcsb_id": objD["rcsb_id"]}
+                updateDL.append({"selectD": selectD, "updateD": objD})
+            except Exception as e:
+                logger.exception("Failing with %s", str(e))
+        obUpd = ObjectUpdater(self.__cfgOb)
+        numUpd = obUpd.update(databaseName, collectionName, updateDL)
+        logger.info("Updated reference count is %d", numUpd)
+
+    def __reload(self, expireDays=0, fetchLimit=None, updateOnReload=True, retryUnmatched=False, **kwargs):
+        """Reload match index and entry data from object store.  Continue update operations as specified.
+
+        Args:
+            expireDays (int): expiration days on match data (default 0 meaning none)
+            fetchLimit (int): limit to the number of entry updates performed (None)
+            updateOnReload (bool): update any definitions in the search not match index (default=True)
+            retryUnMatch (bool): update includes previously unmatched search definitions with update (default=False)
+
+        Returns:
+            (dict,dict,obj,obj): match index dictionary, reference data dictionary,
+                                 chemical component index provider, chemical component search index provider
+
 
 
         ChemicalIdentifierFields = ("idCode", "identifierSource", "identifierType", "identifier")
@@ -287,35 +404,37 @@ class PubChemCacheProvider(object):
             # Get current the indices of source chemical reference data -
             ok, ccidxP, ccsidxP = self.__rebuildChemCompSourceIndices(**kwargs)
             if not ok:
-                return matchD, refD
-            ccIdxD = ccidxP.getIndex()
-            # Target list of local chemical component and BIRD identifiers
-            sourceIdList = sorted(ccIdxD.keys())
-            if self.__useCache:
-                logger.info("Reloading chemical reference data (expireDays %r)", expireDays)
-                matchedIdList = self.__getReferenceDataIds(expireDays=expireDays)
-            # --
-            logger.info("With useCache %r matched reference identifier count (%d) ", self.__useCache, len(matchedIdList))
-            updateIdList = sorted(set(sourceIdList) - set(matchedIdList))
-            logger.info("Missing chemical definition correspondences %d fetchLimit %r", len(updateIdList), fetchLimit)
+                return matchD, refD, ccidxP, ccsidxP
+            if updateOnReload:
+                ccIdxD = ccidxP.getIndex()
+                # Index of target of local chemical component and BIRD identifiers
+                sourceIdList = sorted(ccIdxD.keys())
+                if self.__useCache:
+                    filterMatched = True if retryUnmatched else False
+                    logger.info("Reloading chemical reference data (expireDays %r)", expireDays)
+                    matchedIdList = self.__getReferenceDataIds(expireDays=expireDays, filterMatched=filterMatched)
+                # --
+                logger.info("With useCache %r matched reference identifier count (%d) ", self.__useCache, len(matchedIdList))
+                updateIdList = sorted(set(sourceIdList) - set(matchedIdList))
+                logger.info("Missing chemical definition correspondences %d fetchLimit %r", len(updateIdList), fetchLimit)
+                #
+                updateIdList = updateIdList[:fetchLimit] if fetchLimit else updateIdList
+                #
+                if updateIdList:
+                    searchIdxD = ccsidxP.getIndex()
+                    logger.info("Update reference data cache for %d chemical identifers", len(updateIdList))
+                    ok, failList = self.__updateReferenceData(updateIdList, searchIdxD, **kwargs)
+                    logger.info("Update reference data status is %r missing count %d", ok, len(failList))
+                else:
+                    logger.info("No reference data updates required")
             #
-            updateIdList = updateIdList[:fetchLimit] if fetchLimit else updateIdList
-            #
-            if updateIdList:
-                searchIdxD = ccsidxP.getIndex()
-                logger.info("Update reference data cache for %d chemical identifers", len(updateIdList))
-                ok, failList = self.__updateReferenceData(updateIdList, searchIdxD, **kwargs)
-                logger.info("Update reference data status is %r missing count %d", ok, len(failList))
-            else:
-                logger.info("No reference data updates required")
-            #
-            matchD = self.__getReferenceData(self.__databaseName, self.__matchDataCollectionName)
+            matchD = self.__getReferenceData(self.__databaseName, self.__matchIndexCollectionName)
             refD = self.__getReferenceData(self.__databaseName, self.__refDataCollectionName)
             logger.info("Completed - returning match length %d and reference data length %d", len(matchD), len(refD))
-            return matchD, refD
+            return matchD, refD, ccidxP, ccsidxP
         except Exception as e:
             logger.exception("Failing with %s", str(e))
-        return None, None
+        return None, None, None, None
 
     def __updateReferenceData(self, idList, searchIdxD, **kwargs):
         """Launch worker methods to update chemical reference data correspondences.
@@ -344,8 +463,9 @@ class PubChemCacheProvider(object):
     def __rebuildChemCompSourceIndices(self, **kwargs):
         """ Rebuild source indices of chemical component definitions.
         """
-
+        logger.info("Rebuilding chemical definition index.")
         ok1, ccidxP = self.__buildChemCompIndex(**kwargs)
+        logger.info("Rebuilding chemical search indices.")
         ok2, ccsidxP = self.__buildChemCompSearchIndex(**kwargs)
         return ok1 & ok2, ccidxP, ccsidxP
 
